@@ -15,29 +15,35 @@
 """
 __docformat__ = 'restructuredtext'
 
+try:
+    from collections import abc
+except ImportError: # pragma: no cover
+    # Python 2
+    import collections as abc
+
 from datetime import datetime
 from datetime import date
 from datetime import timedelta
 from datetime import time
 import decimal
 import re
-import threading
 
-from zope.event import notify
+
 from zope.interface import classImplements
 from zope.interface import implementer
-from zope.interface import Interface
 from zope.interface.interfaces import IInterface
-from zope.interface.interfaces import IMethod
+
 
 from zope.schema.interfaces import IASCII
 from zope.schema.interfaces import IASCIILine
 from zope.schema.interfaces import IBaseVocabulary
-from zope.schema.interfaces import IBeforeObjectAssignedEvent
 from zope.schema.interfaces import IBool
 from zope.schema.interfaces import IBytes
 from zope.schema.interfaces import IBytesLine
 from zope.schema.interfaces import IChoice
+from zope.schema.interfaces import ICollection
+from zope.schema.interfaces import IComplex
+from zope.schema.interfaces import IContainer
 from zope.schema.interfaces import IContextSourceBinder
 from zope.schema.interfaces import IDate
 from zope.schema.interfaces import IDatetime
@@ -49,13 +55,22 @@ from zope.schema.interfaces import IFloat
 from zope.schema.interfaces import IFromUnicode
 from zope.schema.interfaces import IFrozenSet
 from zope.schema.interfaces import IId
+from zope.schema.interfaces import IIterable
 from zope.schema.interfaces import IInt
+from zope.schema.interfaces import IIntegral
 from zope.schema.interfaces import IInterfaceField
 from zope.schema.interfaces import IList
 from zope.schema.interfaces import IMinMaxLen
+from zope.schema.interfaces import IMapping
+from zope.schema.interfaces import IMutableMapping
+from zope.schema.interfaces import IMutableSequence
 from zope.schema.interfaces import IObject
+from zope.schema.interfaces import INumber
 from zope.schema.interfaces import IPassword
+from zope.schema.interfaces import IReal
+from zope.schema.interfaces import IRational
 from zope.schema.interfaces import ISet
+from zope.schema.interfaces import ISequence
 from zope.schema.interfaces import ISource
 from zope.schema.interfaces import ISourceText
 from zope.schema.interfaces import IText
@@ -70,14 +85,13 @@ from zope.schema.interfaces import InvalidValue
 from zope.schema.interfaces import WrongType
 from zope.schema.interfaces import WrongContainedType
 from zope.schema.interfaces import NotUnique
-from zope.schema.interfaces import SchemaNotProvided
-from zope.schema.interfaces import SchemaNotFullyImplemented
 from zope.schema.interfaces import InvalidURI
 from zope.schema.interfaces import InvalidId
 from zope.schema.interfaces import InvalidDottedName
 from zope.schema.interfaces import ConstraintNotSatisfied
 
 from zope.schema._bootstrapfields import Field
+from zope.schema._bootstrapfields import Complex
 from zope.schema._bootstrapfields import Container  # API import for __init__
 from zope.schema._bootstrapfields import Iterable
 from zope.schema._bootstrapfields import Orderable
@@ -85,11 +99,16 @@ from zope.schema._bootstrapfields import Text
 from zope.schema._bootstrapfields import TextLine
 from zope.schema._bootstrapfields import Bool
 from zope.schema._bootstrapfields import Int
+from zope.schema._bootstrapfields import Integral
+from zope.schema._bootstrapfields import Number
 from zope.schema._bootstrapfields import Password
+from zope.schema._bootstrapfields import Rational
+from zope.schema._bootstrapfields import Real
 from zope.schema._bootstrapfields import MinMaxLen
+from zope.schema._bootstrapfields import _NotGiven
+from zope.schema._bootstrapfields import Object
 from zope.schema.fieldproperty import FieldProperty
 from zope.schema.vocabulary import getVocabularyRegistry
-from zope.schema.vocabulary import VocabularyRegistryError
 from zope.schema.vocabulary import SimpleVocabulary
 
 
@@ -99,9 +118,6 @@ from zope.schema._compat import binary_type
 from zope.schema._compat import PY3
 from zope.schema._compat import make_binary
 
-# pep 8 friendlyness
-Container
-
 # Fix up bootstrap field types
 Field.title = FieldProperty(IField['title'])
 Field.description = FieldProperty(IField['description'])
@@ -118,8 +134,19 @@ classImplements(TextLine, ITextLine)
 classImplements(Password, IPassword)
 classImplements(Bool, IBool)
 classImplements(Bool, IFromUnicode)
+classImplements(Iterable, IIterable)
+classImplements(Container, IContainer)
+
+classImplements(Number, INumber)
+classImplements(Complex, IComplex)
+classImplements(Real, IReal)
+classImplements(Rational, IRational)
+classImplements(Integral, IIntegral)
 classImplements(Int, IInt)
 
+classImplements(Object, IObject)
+
+
 
 @implementer(ISourceText)
 class SourceText(Text):
@@ -156,7 +183,7 @@ class ASCII(NativeString):
         if not value:
             return
         if not max(map(ord, value)) < 128:
-            raise InvalidValue
+            raise InvalidValue().with_field_and_value(self, value)
 
 
 @implementer(IBytesLine)
@@ -183,39 +210,85 @@ class ASCIILine(ASCII):
         return '\n' not in value
 
 
+class InvalidFloatLiteral(ValueError, ValidationError):
+    """Raised by Float fields."""
+
+
 @implementer(IFloat, IFromUnicode)
-class Float(Orderable, Field):
-    __doc__ = IFloat.__doc__
+class Float(Real):
+    """
+    A field representing a native :class:`float` and implementing
+    :class:`zope.schema.interfaces.IFloat`.
+
+    The class :class:`zope.schema.Real` is a more general version,
+    accepting floats, integers, and fractions.
+
+    The :meth:`fromUnicode` method only accepts values that can be parsed
+    by the ``float`` constructor::
+
+        >>> from zope.schema._field import Float
+        >>> f = Float()
+        >>> f.fromUnicode("1")
+        1.0
+        >>> f.fromUnicode("125.6")
+        125.6
+        >>> f.fromUnicode("1+0j") # doctest: +IGNORE_EXCEPTION_DETAIL
+        Traceback (most recent call last):
+        ...
+        InvalidFloatLiteral: Invalid literal for float(): 1+0j
+        >>> f.fromUnicode("1/2") # doctest: +IGNORE_EXCEPTION_DETAIL
+        Traceback (most recent call last):
+        ...
+        InvalidFloatLiteral: invalid literal for float(): 1/2
+        >>> f.fromUnicode(str(2**31234) + '.' + str(2**256)) # doctest: +ELLIPSIS
+        inf
+        >>> f.fromUnicode("not a number") # doctest: +IGNORE_EXCEPTION_DETAIL
+        Traceback (most recent call last):
+        ...
+        InvalidFloatLiteral: could not convert string to float: not a number
+    """
     _type = float
+    _unicode_converters = (float,)
+    _validation_error = InvalidFloatLiteral
 
-    def __init__(self, *args, **kw):
-        super(Float, self).__init__(*args, **kw)
 
-    def fromUnicode(self, uc):
-        """ See IFromUnicode.
-        """
-        v = float(uc)
-        self.validate(v)
-        return v
+class InvalidDecimalLiteral(ValueError, ValidationError):
+    "Raised by decimal fields"
 
 
 @implementer(IDecimal, IFromUnicode)
-class Decimal(Orderable, Field):
-    __doc__ = IDecimal.__doc__
+class Decimal(Number):
+    """
+    A field representing a native :class:`decimal.Decimal` and implementing
+    :class:`zope.schema.interfaces.IDecimal`.
+
+    The :meth:`fromUnicode` method only accepts values that can be parsed
+    by the ``Decimal`` constructor::
+
+        >>> from zope.schema._field import Decimal
+        >>> f = Decimal()
+        >>> f.fromUnicode("1")
+        Decimal('1')
+        >>> f.fromUnicode("125.6")
+        Decimal('125.6')
+        >>> f.fromUnicode("1+0j") # doctest: +IGNORE_EXCEPTION_DETAIL
+        Traceback (most recent call last):
+        ...
+        InvalidDecimalLiteral: Invalid literal for Decimal(): 1+0j
+        >>> f.fromUnicode("1/2") # doctest: +IGNORE_EXCEPTION_DETAIL
+        Traceback (most recent call last):
+        ...
+        InvalidDecimalLiteral: Invalid literal for Decimal(): 1/2
+        >>> f.fromUnicode(str(2**31234) + '.' + str(2**256)) # doctest: +ELLIPSIS
+        Decimal('2349...936')
+        >>> f.fromUnicode("not a number") # doctest: +IGNORE_EXCEPTION_DETAIL
+        Traceback (most recent call last):
+        ...
+        InvalidDecimalLiteral: could not convert string to float: not a number
+    """
     _type = decimal.Decimal
-
-    def __init__(self, *args, **kw):
-        super(Decimal, self).__init__(*args, **kw)
-
-    def fromUnicode(self, uc):
-        """ See IFromUnicode.
-        """
-        try:
-            v = decimal.Decimal(uc)
-        except decimal.InvalidOperation:
-            raise ValueError('invalid literal for Decimal(): %s' % uc)
-        self.validate(v)
-        return v
+    _unicode_converters = (decimal.Decimal,)
+    _validation_error = InvalidDecimalLiteral
 
 
 @implementer(IDatetime)
@@ -235,7 +308,7 @@ class Date(Orderable, Field):
     def _validate(self, value):
         super(Date, self)._validate(value)
         if isinstance(value, datetime):
-            raise WrongType(value, self._type, self.__name__)
+            raise WrongType(value, self._type, self.__name__).with_field_and_value(self, value)
 
 
 @implementer(ITimedelta)
@@ -250,6 +323,22 @@ class Time(Orderable, Field):
     _type = time
 
 
+class MissingVocabularyError(ValidationError,
+                             ValueError,
+                             LookupError):
+    """Raised when a named vocabulary cannot be found."""
+    # Subclasses ValueError and LookupError for backwards compatibility
+
+class InvalidVocabularyError(ValidationError,
+                             ValueError,
+                             TypeError):
+    """Raised when the vocabulary is not an ISource."""
+    # Subclasses TypeError and ValueError for backwards compatibility
+
+    def __init__(self, vocabulary):
+        super(InvalidVocabularyError, self).__init__("Invalid vocabulary %r" % (vocabulary,))
+
+
 @implementer(IChoice, IFromUnicode)
 class Choice(Field):
     """Choice fields can have a value found in a constant or dynamic set of
@@ -287,7 +376,7 @@ class Choice(Field):
         else:
             if (not ISource.providedBy(vocabulary)
                     and not IContextSourceBinder.providedBy(vocabulary)):
-                raise ValueError('Invalid vocabulary')
+                raise InvalidVocabularyError(vocabulary)
             self.vocabulary = vocabulary
         # Before a default value is checked, it is validated. However, a
         # named vocabulary is usually not complete when these fields are
@@ -301,19 +390,34 @@ class Choice(Field):
 
     source = property(lambda self: self.vocabulary)
 
-    def bind(self, object):
-        """See zope.schema._bootstrapinterfaces.IField."""
-        clone = super(Choice, self).bind(object)
-        # get registered vocabulary if needed:
-        if IContextSourceBinder.providedBy(self.vocabulary):
-            clone.vocabulary = self.vocabulary(object)
-        elif clone.vocabulary is None and self.vocabularyName is not None:
+    def _resolve_vocabulary(self, value):
+        # Find the vocabulary we should use, raising
+        # an exception if this isn't possible, and returning
+        # an ISource otherwise.
+        vocabulary = self.vocabulary
+        if IContextSourceBinder.providedBy(vocabulary) and self.context is not None:
+            vocabulary = vocabulary(self.context)
+        elif vocabulary is None and self.vocabularyName is not None:
             vr = getVocabularyRegistry()
-            clone.vocabulary = vr.get(object, self.vocabularyName)
+            try:
+                vocabulary = vr.get(self.context, self.vocabularyName)
+            except LookupError:
+                raise MissingVocabularyError(
+                    "Can't validate value without vocabulary named %r" % (self.vocabularyName,)
+                ).with_field_and_value(self, value)
 
-        if not ISource.providedBy(clone.vocabulary):
-            raise ValueError('Invalid clone vocabulary')
+        if not ISource.providedBy(vocabulary):
+            raise InvalidVocabularyError(vocabulary).with_field_and_value(self, value)
 
+        return vocabulary
+
+    def bind(self, context):
+        """See zope.schema._bootstrapinterfaces.IField."""
+        clone = super(Choice, self).bind(context)
+        # Eagerly get registered vocabulary if needed;
+        # once that's done, just return it
+        vocabulary = clone.vocabulary = clone._resolve_vocabulary(None)
+        clone._resolve_vocabulary = lambda value: vocabulary
         return clone
 
     def fromUnicode(self, str):
@@ -327,15 +431,9 @@ class Choice(Field):
         if self._init_field:
             return
         super(Choice, self)._validate(value)
-        vocabulary = self.vocabulary
-        if vocabulary is None:
-            vr = getVocabularyRegistry()
-            try:
-                vocabulary = vr.get(None, self.vocabularyName)
-            except VocabularyRegistryError:
-                raise ValueError("Can't validate value without vocabulary")
+        vocabulary = self._resolve_vocabulary(value)
         if value not in vocabulary:
-            raise ConstraintNotSatisfied(value, self.__name__)
+            raise ConstraintNotSatisfied(value, self.__name__).with_field_and_value(self, value)
 
 
 _isuri = r"[a-zA-z0-9+.-]+:"  # scheme
@@ -353,7 +451,7 @@ class URI(NativeStringLine):
         if _isuri(value):
             return
 
-        raise InvalidURI(value)
+        raise InvalidURI(value).with_field_and_value(self, value)
 
     def fromUnicode(self, value):
         """ See IFromUnicode.
@@ -370,13 +468,35 @@ _isdotted = re.compile(
     r"$").match
 
 
+class _StrippedNativeStringLine(NativeStringLine):
+
+    _invalid_exc_type = None
+
+    def fromUnicode(self, value):
+        v = value.strip()
+        # On Python 2, self._type is bytes, so we need to encode
+        # unicode down to ASCII bytes. On Python 3, self._type is
+        # unicode, but we don't want to allow non-ASCII values, to match
+        # Python 2 (our regexs would reject that anyway.)
+        try:
+            v = v.encode('ascii') # bytes
+        except UnicodeEncodeError:
+            raise self._invalid_exc_type(value).with_field_and_value(self, value)
+        if not isinstance(v, self._type):
+            v = v.decode('ascii')
+        self.validate(v)
+        return v
+
+
 @implementer(IDottedName)
-class DottedName(NativeStringLine):
+class DottedName(_StrippedNativeStringLine):
     """Dotted name field.
 
     Values of DottedName fields must be Python-style dotted names.
     """
 
+    _invalid_exc_type = InvalidDottedName
+
     def __init__(self, *args, **kw):
         self.min_dots = int(kw.pop("min_dots", 0))
         if self.min_dots < 0:
@@ -394,31 +514,27 @@ class DottedName(NativeStringLine):
         """
         super(DottedName, self)._validate(value)
         if not _isdotted(value):
-            raise InvalidDottedName(value)
+            raise InvalidDottedName(value).with_field_and_value(self, value)
         dots = value.count(".")
         if dots < self.min_dots:
             raise InvalidDottedName(
                 "too few dots; %d required" % self.min_dots, value
-            )
+            ).with_field_and_value(self, value)
         if self.max_dots is not None and dots > self.max_dots:
             raise InvalidDottedName("too many dots; no more than %d allowed" %
-                                    self.max_dots, value)
+                                    self.max_dots, value).with_field_and_value(self, value)
 
-    def fromUnicode(self, value):
-        v = value.strip()
-        if not isinstance(v, self._type):
-            v = v.encode('ascii')
-        self.validate(v)
-        return v
 
 
 @implementer(IId, IFromUnicode)
-class Id(NativeStringLine):
+class Id(_StrippedNativeStringLine):
     """Id field
 
     Values of id fields must be either uris or dotted names.
     """
 
+    _invalid_exc_type = InvalidId
+
     def _validate(self, value):
         super(Id, self)._validate(value)
         if _isuri(value):
@@ -426,16 +542,7 @@ class Id(NativeStringLine):
         if _isdotted(value) and "." in value:
             return
 
-        raise InvalidId(value)
-
-    def fromUnicode(self, value):
-        """ See IFromUnicode.
-        """
-        v = value.strip()
-        if not isinstance(v, self._type):
-            v = v.encode('ascii')
-        self.validate(v)
-        return v
+        raise InvalidId(value).with_field_and_value(self, value)
 
 
 @implementer(IInterfaceField)
@@ -445,7 +552,11 @@ class InterfaceField(Field):
     def _validate(self, value):
         super(InterfaceField, self)._validate(value)
         if not IInterface.providedBy(value):
-            raise WrongType("An interface is required", value, self.__name__)
+            raise WrongType(
+                "An interface is required",
+                value,
+                self.__name__
+            ).with_field_and_value(self, value)
 
 
 def _validate_sequence(value_type, value, errors=None):
@@ -463,9 +574,10 @@ def _validate_sequence(value_type, value, errors=None):
 
     To validate a sequence of various values:
 
-       >>> errors = _validate_sequence(field, (b'foo', u'bar', 1))
-       >>> errors # XXX assumes Python2 reprs
-       [WrongType('foo', <type 'unicode'>, ''), WrongType(1, <type 'unicode'>, '')]
+       >>> from zope.schema._compat import text_type
+       >>> errors = _validate_sequence(field, (bytearray(b'foo'), u'bar', 1))
+       >>> errors
+       [WrongType(bytearray(b'foo'), <...>, ''), WrongType(1, <...>, '')]
 
     The only valid value in the sequence is the second item. The others
     generated errors.
@@ -474,8 +586,8 @@ def _validate_sequence(value_type, value, errors=None):
     for a new sequence:
 
        >>> errors = _validate_sequence(field, (2, u'baz'), errors)
-       >>> errors # XXX assumes Python2 reprs
-       [WrongType('foo', <type 'unicode'>, ''), WrongType(1, <type 'unicode'>, ''), WrongType(2, <type 'unicode'>, '')]
+       >>> errors
+       [WrongType(bytearray(b'foo'), <...>, ''), WrongType(1, <...>, ''), WrongType(2, <...>, '')]
 
     """
     if errors is None:
@@ -490,180 +602,141 @@ def _validate_sequence(value_type, value, errors=None):
     return errors
 
 
-def _validate_uniqueness(value):
+def _validate_uniqueness(self, value):
     temp_values = []
     for item in value:
         if item in temp_values:
-            raise NotUnique(item)
+            raise NotUnique(item).with_field_and_value(self, value)
 
         temp_values.append(item)
 
 
-class AbstractCollection(MinMaxLen, Iterable):
+@implementer(ICollection)
+class Collection(MinMaxLen, Iterable):
+    """
+    A generic collection implementing :class:`zope.schema.interfaces.ICollection`.
+
+    Subclasses can define the attribute ``value_type`` to be a field
+    such as an :class:`Object` that will be checked for each member of
+    the collection. This can then be omitted from the constructor call.
+
+    They can also define the attribute ``_type`` to be a concrete
+    class (or tuple of classes) that the collection itself will
+    be checked to be an instance of. This cannot be set in the constructor.
+
+    .. versionchanged:: 4.6.0
+       Add the ability for subclasses to specify ``value_type``
+       and ``unique``, and allow eliding them from the constructor.
+    """
     value_type = None
     unique = False
 
-    def __init__(self, value_type=None, unique=False, **kw):
-        super(AbstractCollection, self).__init__(**kw)
+    def __init__(self, value_type=_NotGiven, unique=_NotGiven, **kw):
+        super(Collection, self).__init__(**kw)
         # whine if value_type is not a field
-        if value_type is not None and not IField.providedBy(value_type):
-            raise ValueError("'value_type' must be field instance.")
-        self.value_type = value_type
-        self.unique = unique
+        if value_type is not _NotGiven:
+            self.value_type = value_type
 
-    def bind(self, object):
+        if self.value_type is not None and not IField.providedBy(self.value_type):
+            raise ValueError("'value_type' must be field instance.")
+        if unique is not _NotGiven:
+            self.unique = unique
+
+    def bind(self, context):
         """See zope.schema._bootstrapinterfaces.IField."""
-        clone = super(AbstractCollection, self).bind(object)
+        clone = super(Collection, self).bind(context)
         # binding value_type is necessary for choices with named vocabularies,
         # and possibly also for other fields.
         if clone.value_type is not None:
-            clone.value_type = clone.value_type.bind(object)
+            clone.value_type = clone.value_type.bind(context)
         return clone
 
     def _validate(self, value):
-        super(AbstractCollection, self)._validate(value)
+        super(Collection, self)._validate(value)
         errors = _validate_sequence(self.value_type, value)
         if errors:
-            raise WrongContainedType(errors, self.__name__)
+            try:
+                raise WrongContainedType(errors, self.__name__).with_field_and_value(self, value)
+            finally:
+                # Break cycles
+                del errors
         if self.unique:
-            _validate_uniqueness(value)
+            _validate_uniqueness(self, value)
+
+
+#: An alternate name for :class:`.Collection`.
+#:
+#: .. deprecated:: 4.6.0
+#:   Use :class:`.Collection` instead.
+AbstractCollection = Collection
+
+
+@implementer(ISequence)
+class Sequence(Collection):
+    """
+    A field representing an ordered sequence.
+
+    .. versionadded:: 4.6.0
+    """
+    _type = abc.Sequence
 
 
 @implementer(ITuple)
-class Tuple(AbstractCollection):
+class Tuple(Sequence):
     """A field representing a Tuple."""
     _type = tuple
 
 
+@implementer(IMutableSequence)
+class MutableSequence(Sequence):
+    """
+    A field representing a mutable sequence.
+
+    .. versionadded:: 4.6.0
+    """
+    _type = abc.MutableSequence
+
+
 @implementer(IList)
-class List(AbstractCollection):
+class List(MutableSequence):
     """A field representing a List."""
     _type = list
 
 
+class _AbstractSet(Collection):
+    unique = True
+
+    def __init__(self, *args, **kwargs):
+        super(_AbstractSet, self).__init__(*args, **kwargs)
+        if not self.unique: # set members are always unique
+            raise TypeError(
+                "__init__() got an unexpected keyword argument 'unique'")
+
+
 @implementer(ISet)
-class Set(AbstractCollection):
+class Set(_AbstractSet):
     """A field representing a set."""
     _type = set
 
-    def __init__(self, **kw):
-        if 'unique' in kw:  # set members are always unique
-            raise TypeError(
-                "__init__() got an unexpected keyword argument 'unique'")
-        super(Set, self).__init__(unique=True, **kw)
-
 
 @implementer(IFrozenSet)
-class FrozenSet(AbstractCollection):
+class FrozenSet(_AbstractSet):
     _type = frozenset
 
-    def __init__(self, **kw):
-        if 'unique' in kw:  # set members are always unique
-            raise TypeError(
-                "__init__() got an unexpected keyword argument 'unique'")
-        super(FrozenSet, self).__init__(unique=True, **kw)
 
+@implementer(IMapping)
+class Mapping(MinMaxLen, Iterable):
+    """
+    A field representing a mapping.
 
-VALIDATED_VALUES = threading.local()
-
-
-def _validate_fields(schema, value, errors=None):
-    if errors is None:
-        errors = []
-    # Interface can be used as schema property for Object fields that plan to
-    # hold values of any type.
-    # Because Interface does not include any Attribute, it is obviously not
-    # worth looping on its methods and filter them all out.
-    if schema is Interface:
-        return errors
-    # if `value` is part of a cyclic graph, we need to break the cycle to avoid
-    # infinite recursion. Collect validated objects in a thread local dict by
-    # it's python represenation. A previous version was setting a volatile
-    # attribute which didn't work with security proxy
-    if id(value) in VALIDATED_VALUES.__dict__:
-        return errors
-    VALIDATED_VALUES.__dict__[id(value)] = True
-    # (If we have gotten here, we know that `value` provides an interface
-    # other than zope.interface.Interface;
-    # iow, we can rely on the fact that it is an instance
-    # that supports attribute assignment.)
-    try:
-        for name in schema.names(all=True):
-            if not IMethod.providedBy(schema[name]):
-                try:
-                    attribute = schema[name]
-                    if IChoice.providedBy(attribute):
-                        # Choice must be bound before validation otherwise
-                        # IContextSourceBinder is not iterable in validation
-                        bound = attribute.bind(value)
-                        bound.validate(getattr(value, name))
-                    elif IField.providedBy(attribute):
-                        # validate attributes that are fields
-                        attribute.validate(getattr(value, name))
-                except ValidationError as error:
-                    errors.append(error)
-                except AttributeError as error:
-                    # property for the given name is not implemented
-                    errors.append(SchemaNotFullyImplemented(error))
-    finally:
-        del VALIDATED_VALUES.__dict__[id(value)]
-    return errors
-
-
-@implementer(IObject)
-class Object(Field):
-    __doc__ = IObject.__doc__
-
-    def __init__(self, schema, **kw):
-        if not IInterface.providedBy(schema):
-            raise WrongType
-
-        self.schema = schema
-        super(Object, self).__init__(**kw)
-
-    def _validate(self, value):
-        super(Object, self)._validate(value)
-
-        # schema has to be provided by value
-        if not self.schema.providedBy(value):
-            raise SchemaNotProvided
-
-        # check the value against schema
-        errors = _validate_fields(self.schema, value)
-        if errors:
-            raise WrongContainedType(errors, self.__name__)
-
-    def set(self, object, value):
-        # Announce that we're going to assign the value to the object.
-        # Motivation: Widgets typically like to take care of policy-specific
-        # actions, like establishing location.
-        event = BeforeObjectAssignedEvent(value, self.__name__, object)
-        notify(event)
-        # The event subscribers are allowed to replace the object, thus we need
-        # to replace our previous value.
-        value = event.object
-        super(Object, self).set(object, value)
-
-
-@implementer(IBeforeObjectAssignedEvent)
-class BeforeObjectAssignedEvent(object):
-    """An object is going to be assigned to an attribute on another object."""
-
-    def __init__(self, object, name, context):
-        self.object = object
-        self.name = name
-        self.context = context
-
-
-@implementer(IDict)
-class Dict(MinMaxLen, Iterable):
-    """A field representing a Dict."""
-    _type = dict
+    .. versionadded:: 4.6.0
+    """
+    _type = abc.Mapping
     key_type = None
     value_type = None
 
     def __init__(self, key_type=None, value_type=None, **kw):
-        super(Dict, self).__init__(**kw)
+        super(Mapping, self).__init__(**kw)
         # whine if key_type or value_type is not a field
         if key_type is not None and not IField.providedBy(key_type):
             raise ValueError("'key_type' must be field instance.")
@@ -673,23 +746,23 @@ class Dict(MinMaxLen, Iterable):
         self.value_type = value_type
 
     def _validate(self, value):
-        super(Dict, self)._validate(value)
+        super(Mapping, self)._validate(value)
         errors = []
-        try:
-            if self.value_type:
-                errors = _validate_sequence(self.value_type, value.values(),
-                                            errors)
-            errors = _validate_sequence(self.key_type, value, errors)
+        if self.value_type:
+            errors = _validate_sequence(self.value_type, value.values(),
+                                        errors)
+        errors = _validate_sequence(self.key_type, value, errors)
 
-            if errors:
-                raise WrongContainedType(errors, self.__name__)
-
-        finally:
-            errors = None
+        if errors:
+            try:
+                raise WrongContainedType(errors, self.__name__).with_field_and_value(self, value)
+            finally:
+                # Break cycles
+                del errors
 
     def bind(self, object):
         """See zope.schema._bootstrapinterfaces.IField."""
-        clone = super(Dict, self).bind(object)
+        clone = super(Mapping, self).bind(object)
         # binding value_type is necessary for choices with named vocabularies,
         # and possibly also for other fields.
         if clone.key_type is not None:
@@ -697,3 +770,19 @@ class Dict(MinMaxLen, Iterable):
         if clone.value_type is not None:
             clone.value_type = clone.value_type.bind(object)
         return clone
+
+
+@implementer(IMutableMapping)
+class MutableMapping(Mapping):
+    """
+    A field representing a mutable mapping.
+
+    .. versionadded:: 4.6.0
+    """
+    _type = abc.MutableMapping
+
+
+@implementer(IDict)
+class Dict(MutableMapping):
+    """A field representing a Dict."""
+    _type = dict
